/************************************************************
 **
 ** picmovlpop.c : PIC data optimization definitions
 **
 ** Copyright (c) 2010, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include "../libcore/pf_proc.h"
#include "pic_msg.h"
#include "piccolst.h"
#include "pic_brop.h"
#include "picmovlpop.h"

/*
 * NAME
 *   pic_code_movlp_analyze
 *
 * DESCRIPTION
 *   analyze the movlp in the pic_codes[] array
 *
 * PARAMETERS
 *   pf    : 
 *   pcl   : pcl on entry
 *   pclath: current contents of pclath
 *   depth : depth (for statistics only)
 *
 * RETURN
 *   none
 *
 * NOTES
 */
static void pic_code_movlp_analyze(pfile_t *pf, pic_code_t code,
  uchar pclath, unsigned depth)
{
  boolean_t pv_code_is_cond;

  /* now, traverse the code until either
     (1) all bits get known values
     (2) we hit an unconditional return or a leave */
  pv_code_is_cond = boolean_false;

  do {
    boolean_t  code_is_cond;
    pic_code_t next;

    code_is_cond = boolean_false;
    next = pic_code_next_get(code);

    if ((pic_code_pclath_get(code) == pclath) 
      || (PIC_PCLATH_INDETERMINITE == pic_code_pclath_get(code))) {
      next = 0; /* we've a match, we're done! */
    } else {
      if (PIC_PCLATH_UNKNOWN == pic_code_pclath_get(code)) {
        pic_code_pclath_set(code, pclath);
      } else if (pclath != pic_code_pclath_get(code)) {
        pic_code_pclath_set(code, PIC_PCLATH_INDETERMINITE);
      }
      switch (pic_code_op_get(code)) {
        case pic_opcode_movlp:
          {
            label_t lbl;

            lbl = pic_code_brdst_get(code);
            pclath = label_pc_get(lbl) >> 8;
          }
          break;
        case pic_opcode_return:
        case pic_opcode_retfie:
        case pic_opcode_retlw:
          /* if this is unconditional we're done otherwise simply continue */
          if (!pv_code_is_cond) {
            next = 0;
          }
          break;
        case pic_opcode_goto:
        case pic_opcode_call:
          {
            boolean_t  is_suspend;
            pic_code_t brcode;

            brcode = pic_code_label_find(pf, pic_code_brdst_get(code));
            is_suspend = pic_code_is_suspend(pf, code);

            if ((pic_opcode_goto == pic_code_op_get(code))
                && !pv_code_is_cond
                && !is_suspend) {
              /* an absolute goto so next = brcode */
              next = brcode;
            } else {
              /* this is either a conditional goto or a call. either
                 way we first process then branch, then continue
                 processing from where we were. for the time being
                 a call is going to invalidate the known bits */
              pic_code_movlp_analyze(pf, brcode, pclath, depth + 1);
              if (is_suspend || (pic_opcode_call == pic_code_op_get(code))) {
                pclath = PIC_PCLATH_INDETERMINITE;
              }
            }
          }
          break;
        case pic_opcode_incfsz:
        case pic_opcode_decfsz:
        case pic_opcode_btfsc:
        case pic_opcode_btfss:
          code_is_cond = boolean_true;
          break;
        default:
          break;
      }
      if (pic_code_modifies_pcl(pf, code)) {
        next = 0;
      }
    }
    pv_code_is_cond = code_is_cond;
    code = next;
  } while (code);
}

/* run through the pic_codes and remove any redundant dataset commands */
static unsigned pic_code_pclath_post_analyze(pfile_t *pf)
{
  pic_code_t pv;
  pic_code_t code;
  unsigned   ins_ct; /* instruction ct */
  unsigned   rem_ct;

  for (ins_ct = 0, rem_ct = 0, pv = 0, code = pic_code_list_head_get(pf);
       code;
       ins_ct++) {
    pic_code_t next;

    next = pic_code_next_get(code);
    if (!pic_code_flag_test(code, PIC_CODE_FLAG_NO_OPTIMIZE)) {
      boolean_t            cremove;

      switch (pic_code_op_get(code)) {
        case pic_opcode_movlp:
          cremove = (label_pc_get(pic_code_brdst_get(code)) >> 8)
              == pic_code_pclath_get(code);
          break;
        default:
          cremove = boolean_false;
          break;
      }
      if (cremove) {
        pic_code_list_remove(pf, code);
        pic_code_free(code);
        rem_ct++;
      } else {
        pv = code;
      }
    }
    code = next;
  }
  pfile_log(pf, pfile_log_debug, "...removed %u of %u instructions", 
      rem_ct, ins_ct);
  return rem_ct;
}

void pic_code_movlp_optimize(pfile_t *pf)
{
  label_t  lbl;

  pfile_log(pf, pfile_log_debug, PIC_MSG_FIXING_DATA_BITS);
  /* first -- analyze the ISR bits */
  do {
    pfile_proc_t *proc;

    pic_branchbits_pc_set(pf);
    lbl = pfile_label_find(pf, pfile_log_none, PIC_LABEL_ISR);
    if (lbl) {
      pic_code_t code;

      pfile_log(pf, pfile_log_debug, "...analyzing ISR");
      code = pic_code_label_find(pf, lbl);
      if (code) {
        pic_code_movlp_analyze(pf, code, PIC_PCLATH_INDETERMINITE, 1);
      }
      label_release(lbl);
    }

    /* second -- analyze the USER bits */
    lbl = pfile_label_find(pf, pfile_log_err, PIC_LABEL_RESET);
    if (lbl) {
      pic_code_t code;

      pfile_log(pf, pfile_log_debug, "...analyzing USER");
      code = pic_code_label_find(pf, lbl);
      if (code) {
        pic_code_movlp_analyze(pf, code, 0, 1);
      }
      label_release(lbl);
    }
    /* find all procedures that are called indirectly */
    for (proc = pfile_proc_root_get(pf);
         proc;
         proc = pfile_proc_next(proc)) {
      if (pfile_proc_flag_test(proc, PFILE_PROC_FLAG_INDIRECT)
          || pfile_proc_flag_test(proc, PFILE_PROC_FLAG_TASK)) {
        lbl = pfile_proc_label_get(proc);
        if (lbl) {
          pic_code_t code;

          code = (pic_code_t) label_code_get(lbl);
          if (code) {
             pic_code_movlp_analyze(pf, code, PIC_PCLATH_INDETERMINITE, 1);
          }
        }
      }
    }
  } while (pic_code_pclath_post_analyze(pf));
}

